/**
 *    Copyright (C) 2020-present MongoDB, Inc.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the Server Side Public License, version 1,
 *    as published by MongoDB, Inc.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    Server Side Public License for more details.
 *
 *    You should have received a copy of the Server Side Public License
 *    along with this program. If not, see
 *    <http://www.mongodb.com/licensing/server-side-public-license>.
 *
 *    As a special exception, the copyright holders give permission to link the
 *    code of portions of this program with the OpenSSL library under certain
 *    conditions as described in each individual source file and distribute
 *    linked combinations including the program with the OpenSSL library. You
 *    must comply with the Server Side Public License in all respects for
 *    all of the code used other than as permitted herein. If you modify file(s)
 *    with this exception, you may extend this exception to your version of the
 *    file(s), but you are not obligated to do so. If you do not wish to do so,
 *    delete this exception statement from your version. If you delete this
 *    exception statement from all source files in the program, then also delete
 *    it in the license file.
 */

#include "mongo/platform/basic.h"

#include <string>

#include "mongo/bson/json.h"
#include "mongo/db/cst/bson_lexer.h"
#include "mongo/db/cst/parser_gen.hpp"

#include "mongo/unittest/unittest.h"

namespace mongo {
namespace {

void assertTokensMatch(BSONLexer& lexer, std::vector<ParserGen::token::yytokentype> tokens) {
    for (auto&& token : tokens) {
        ASSERT_EQ(lexer.getNext().type_get(), token);
    }
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::END_OF_FILE);
}

TEST(BSONLexerTest, TokenizesOpaqueUserObjects) {
    auto input = fromjson("{pipeline: [{a: 2, b: '1', c: \"$path\", d: \"$$NOW\"}]}");
    BSONLexer lexer(input["pipeline"]);
    assertTokensMatch(lexer,
                      {ParserGen::token::ARG_PIPELINE,
                       ParserGen::token::START_ARRAY,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::FIELDNAME,
                       ParserGen::token::INT_OTHER,
                       ParserGen::token::FIELDNAME,
                       ParserGen::token::STRING,
                       ParserGen::token::FIELDNAME,
                       ParserGen::token::DOLLAR_STRING,
                       ParserGen::token::FIELDNAME,
                       ParserGen::token::DOLLAR_DOLLAR_STRING,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_ARRAY});
}

TEST(BSONLexerTest, TokenizesReservedKeywords) {
    auto input = fromjson("{pipeline: [{$_internalInhibitOptimization: {}}]}");
    BSONLexer lexer(input["pipeline"]);
    assertTokensMatch(lexer,
                      {ParserGen::token::ARG_PIPELINE,
                       ParserGen::token::START_ARRAY,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::STAGE_INHIBIT_OPTIMIZATION,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_ARRAY});
}

TEST(BSONLexerTest, TokenizesReservedKeywordsAtAnyDepth) {
    auto input = fromjson("{pipeline: [{a: {$_internalInhibitOptimization: {}}}]}");
    BSONLexer lexer(input["pipeline"]);
    assertTokensMatch(lexer,
                      {ParserGen::token::ARG_PIPELINE,
                       ParserGen::token::START_ARRAY,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::FIELDNAME,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::STAGE_INHIBIT_OPTIMIZATION,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_OBJECT,
                       ParserGen::token::END_ARRAY});
}

TEST(BSONLexerTest, MidRuleActionToSortNestedObject) {
    auto input = fromjson("{pipeline: [{pipeline: 2.0, coll: 'test'}]}");
    BSONLexer lexer(input["pipeline"]);
    // Iterate until the first object.
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::ARG_PIPELINE);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_ARRAY);
    // Kick the lexer to sort the object, which should move element 'coll' in front of 'pipeline'.
    // Not that this only works because these are reserved keywords recognized by the lexer,
    // arbitrary string field names with *not* get sorted.
    lexer.sortObjTokens();
    auto expected = {ParserGen::token::START_OBJECT,
                     ParserGen::token::ARG_COLL,
                     ParserGen::token::STRING,
                     ParserGen::token::ARG_PIPELINE,
                     ParserGen::token::DOUBLE_OTHER,
                     ParserGen::token::END_OBJECT,
                     ParserGen::token::END_ARRAY};
    assertTokensMatch(lexer, expected);
}


TEST(BSONLexerTest, MidRuleActionToSortDoesNotSortNestedObjects) {
    auto input = fromjson(
        "{pipeline: [{$unionWith: {pipeline: [{$unionWith: 'inner', a: 3.0}], coll: 'outer'}}]}");
    BSONLexer lexer(input["pipeline"]);
    // Iterate until we reach the $unionWith object.
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::ARG_PIPELINE);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_ARRAY);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_OBJECT);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::STAGE_UNION_WITH);
    lexer.sortObjTokens();
    auto expected = {
        ParserGen::token::START_OBJECT,
        ParserGen::token::ARG_COLL,
        ParserGen::token::STRING,        // coll: 'outer'
        ParserGen::token::ARG_PIPELINE,  // inner pipeline
        ParserGen::token::START_ARRAY,
        ParserGen::token::START_OBJECT,
        // The nested pipeline does *not* get sorted, meaning '$unionWith' stays before 'a'.
        ParserGen::token::STAGE_UNION_WITH,
        ParserGen::token::STRING,  // $unionWith: 'inner'
        ParserGen::token::FIELDNAME,
        ParserGen::token::DOUBLE_OTHER,  // a: 3.0
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
    };
    assertTokensMatch(lexer, expected);
}

TEST(BSONLexerTest, MultipleNestedObjectsAreReorderedCorrectly) {
    auto input = fromjson(
        "{pipeline: [{$unionWith: {pipeline: [{$unionWith: 'inner', a: 3.0}], coll: [{$unionWith: "
        "'innerB', a: 2.0}]}}]}");
    BSONLexer lexer(input["pipeline"]);
    // Iterate until we reach the $unionWith object.
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::ARG_PIPELINE);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_ARRAY);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_OBJECT);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::STAGE_UNION_WITH);
    lexer.sortObjTokens();
    auto expected = {
        ParserGen::token::START_OBJECT,
        ParserGen::token::ARG_COLL,
        ParserGen::token::START_ARRAY,
        ParserGen::token::START_OBJECT,
        // The nested pipeline does *not* get sorted, meaning '$unionWith' stays before 'a'.
        ParserGen::token::STAGE_UNION_WITH,
        ParserGen::token::STRING,        // innerb
        ParserGen::token::FIELDNAME,     // a
        ParserGen::token::DOUBLE_OTHER,  // a: 2.0
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
        // Coll nested object ends here.
        ParserGen::token::ARG_PIPELINE,  // inner pipeline
        ParserGen::token::START_ARRAY,
        ParserGen::token::START_OBJECT,
        // The nested pipeline does *not* get sorted, meaning '$unionWith' stays before 'a'.
        ParserGen::token::STAGE_UNION_WITH,
        ParserGen::token::STRING,        // $unionWith: 'inner'
        ParserGen::token::FIELDNAME,     // a
        ParserGen::token::DOUBLE_OTHER,  // a: 3.0
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
    };
    assertTokensMatch(lexer, expected);
}
TEST(BSONLexerTest, MultiLevelBSONDoesntSortChildren) {
    auto input = fromjson(
        "{pipeline: [{$unionWith: {pipeline: [{$unionWith: {'nested': 3.0, 'apple': 3.0}, a: 3.0}],"
        " coll: 'outer'}}]}");
    BSONLexer lexer(input["pipeline"]);
    // Iterate until we reach the $unionWith object.
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::ARG_PIPELINE);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_ARRAY);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::START_OBJECT);
    ASSERT_EQ(lexer.getNext().type_get(), ParserGen::token::STAGE_UNION_WITH);
    lexer.sortObjTokens();
    auto expected = {
        ParserGen::token::START_OBJECT,
        ParserGen::token::ARG_COLL,
        ParserGen::token::STRING,        // coll: 'outer'
        ParserGen::token::ARG_PIPELINE,  // inner pipeline
        // First nested object
        ParserGen::token::START_ARRAY,
        ParserGen::token::START_OBJECT,
        ParserGen::token::STAGE_UNION_WITH,
        // Second nested object
        ParserGen::token::START_OBJECT,
        ParserGen::token::FIELDNAME,  // nested: 3.0
        ParserGen::token::DOUBLE_OTHER,
        ParserGen::token::FIELDNAME,  // apple: 3.0
        ParserGen::token::DOUBLE_OTHER,
        ParserGen::token::END_OBJECT,
        // End second nested object
        ParserGen::token::FIELDNAME,
        ParserGen::token::DOUBLE_OTHER,  // a: 3.0
        ParserGen::token::END_OBJECT,
        // End first nested object
        ParserGen::token::END_ARRAY,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_OBJECT,
        ParserGen::token::END_ARRAY,
    };
    assertTokensMatch(lexer, expected);
}

TEST(BSONLexerTest, EmptyMatchExpressionsAreLexedCorrectly) {
    BSONLexer lexer(fromjson("{filter: {}}").firstElement());
    assertTokensMatch(lexer,
                      {ParserGen::token::ARG_FILTER,
                       ParserGen::token::START_OBJECT,
                       ParserGen::token::END_OBJECT});
}

TEST(BSONLexerTest, TokenizesObjWithPathCorrectly) {
    auto input = fromjson(
        "{pipeline: [{$project: { m: { $dateToString: { date: '$date', "
        "format: '%Y-%m-%d' } } } } ] }");
    BSONLexer lexer(input["pipeline"]);
    assertTokensMatch(lexer,
                      {
                          ParserGen::token::ARG_PIPELINE,
                          ParserGen::token::START_ARRAY,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::STAGE_PROJECT,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::FIELDNAME,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::DATE_TO_STRING,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::ARG_DATE,
                          ParserGen::token::DOLLAR_STRING,
                          ParserGen::token::ARG_FORMAT,
                          ParserGen::token::STRING,
                          ParserGen::token::END_OBJECT,
                          ParserGen::token::END_OBJECT,
                          ParserGen::token::END_OBJECT,
                          ParserGen::token::END_OBJECT,
                          ParserGen::token::END_ARRAY,
                      });
}

TEST(BSONLexerTest, SortSpecTokensGeneratedCorrectly) {
    auto input = fromjson("{sort: {val: 1, test: -1.0, rand: {$meta: 'textScore'}}}");
    BSONLexer lexer(input["sort"]);
    assertTokensMatch(lexer,
                      {
                          ParserGen::token::ARG_SORT,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::FIELDNAME,
                          ParserGen::token::INT_ONE,
                          ParserGen::token::FIELDNAME,
                          ParserGen::token::DOUBLE_NEGATIVE_ONE,
                          ParserGen::token::FIELDNAME,
                          ParserGen::token::START_OBJECT,
                          ParserGen::token::META,
                          ParserGen::token::TEXT_SCORE,
                          ParserGen::token::END_OBJECT,
                          ParserGen::token::END_OBJECT,
                      });
}

}  // namespace
}  // namespace mongo
